#!/usr/bin/env python

from __future__ import print_function, division

import sys
import argparse
from os.path import join
from math import log10

from dark.dna import compareDNAReads, matchToString
from dark.fasta import FastaReads
from dark.reads import (Reads, addFASTACommandLineOptions,
                        parseFASTACommandLineOptions)
from dark.process import Executor
from dark.utils import parseRangeString


def needle(reads):
    """
    Run a Needleman-Wunsch alignment and return the two sequences.

    @param reads: An iterable of two reads.
    @return: A C{Reads} instance with the two aligned sequences.
    """
    from tempfile import mkdtemp
    from shutil import rmtree

    dir = mkdtemp()

    file1 = join(dir, 'file1.fasta')
    with open(file1, 'w') as fp:
        print(reads[0].toString('fasta'), end='', file=fp)

    file2 = join(dir, 'file2.fasta')
    with open(file2, 'w') as fp:
        print(reads[1].toString('fasta'), end='', file=fp)

    out = join(dir, 'result.fasta')

    Executor().execute("needle -asequence '%s' -bsequence '%s' -auto "
                       "-outfile '%s' -aformat fasta" % (
                           file1, file2, out))

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads(list(FastaReads(out)))
    rmtree(dir)

    return result


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=('Compare two sequences.'))

parser.add_argument(
    '--index1', type=int, default=1,
    help='The (1-based) index in the input of the first sequence.')

parser.add_argument(
    '--index2', type=int, default=2,
    help='The (1-based) index in the input of the second sequence.')

parser.add_argument(
    '--align', default=False, action='store_true',
    help='If given, use needle to do a global alignment of the two sequences.')

parser.add_argument(
    '--alignmentFile',
    help='The file to save the alignment to (implies --align).')

parser.add_argument(
    '--strict', default=False, action='store_true',
    help='If given, do not allow ambiguous nucleotide symbols to match')

parser.add_argument(
    '--sites',
    help=('Specify (1-based) sequence sites to keep. All other sites will '
          'be ignored. The sites must be given in the form e.g., '
          '24,100-200,260.'))

parser.add_argument(
    '--showDiffs', default=False, action='store_true',
    help='Print (1-based) sites where the sequence nucleotides differ.')

addFASTACommandLineOptions(parser)
args = parser.parse_args()

keepSequences = set([args.index1 - 1, args.index2 - 1])

reads = list(parseFASTACommandLineOptions(args).filter(
    keepSequences=keepSequences))

if len(reads) == 1:
    if len(keepSequences) == 1:
        # This is ok, they want to compare a sequence with itself.
        reads = Reads([reads[0], reads[0]])
    else:
        print('Could not find both requested sequence indices. Exiting.')
        sys.exit(1)
elif len(reads) != 2:
    print('Could not find both requested sequence indices. Exiting.')
    sys.exit(1)

if args.alignmentFile:
    args.align = True

if args.align:
    len1, len2 = map(len, reads)
    if len1 == len2:
        print('Pre-alignment, sequence lengths were identical: %s' % len1)
    else:
        print('Pre-alignment, sequence lengths: %d, %d (difference %d)' % (
            len1, len2, abs(len1 - len2)))

    # Align.
    reads = needle(reads)

    if args.alignmentFile:
        assert reads.save(args.alignmentFile) == 2

offsets = (parseRangeString(args.sites, convertToZeroBased=True)
           if args.sites else None)

read1, read2 = reads
len1, len2 = map(len, reads)
identicalLengths = len1 == len2

# Sanity check.
if args.align:
    assert identicalLengths

match = compareDNAReads(read1, read2, matchAmbiguous=(not args.strict),
                        offsets=offsets)

x = 'Post-alignment, sequence' if args.align else 'Sequence'
if identicalLengths:
    print('%s lengths are identical: %s' % (x, len1))
else:
    print('%s lengths: %d, %d (difference %d)' % (x, len1, len2,
                                                  abs(len1 - len2)))

print(matchToString(match, read1, read2, matchAmbiguous=(not args.strict),
                    offsets=offsets))

if args.showDiffs:
    # Print all sites where the sequences differ.
    width = int(log10(max(len1, len2))) + 1
    headerPrinted = False
    for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence),
                                  start=1):
        if a != b:
            if not headerPrinted:
                print('Differences (site, %s, %s):' % (read1.id, read2.id))
                headerPrinted = True
            print('  %*d %s %s' % (width, site, a, b))

    if not headerPrinted:
        print('No sequence differences found.')



#!/usr/bin/env python

from __future__ import print_function

from Bio import SeqIO
import sys

if len(sys.argv) < 2:
    print('Usage: %s file.sff > file.fastq' % sys.argv[0], file=sys.stderr)
    sys.exit(1)
else:
    SeqIO.convert(sys.argv[1], 'sff', sys.stdout, 'fastq')



#!/usr/bin/env python

"""
Read DNA FASTA from stdin and print FASTA to stdout, excluding sequences
with too many bases in low complexity regions.

You must have dustmasker in your shell's PATH for this code to succeed.
dustmasker is part of the NCBI BLAST+ suite. You can get it from
ftp://ftp.ncbi.nlm.nih.gov/blast/executables/blast+/LATEST
"""

from __future__ import print_function

import sys
from os import close, unlink
from time import time
import argparse
from subprocess import call
from tempfile import mkstemp

from dark.fasta import FastaReads


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Filter FASTA for low complexity sequences',
        epilog='Read DNA FASTA from stdin and print FASTA to stdout, '
        'excluding sequences of low complexity.')

    parser.add_argument(
        'fasta', help='The FASTA file to read.')

    parser.add_argument(
        '--maxLowComplexity', metavar='FRAC', type=float, default=0.1,
        help='Sequences with a fraction of low-complexity bases up to this '
        'level will be filtered out.')

    parser.add_argument(
        '--rejectFile', metavar='FILE', default=None,
        help='The name of a file to save rejected sequences to.')

    args = parser.parse_args()
    fd, outFilename = mkstemp()
    close(fd)

    if args.rejectFile is not None:
        rejectFp = open(args.rejectFile, 'w')
        reject = rejectFp.write
    else:
        def reject(_):
            return None

    cmd = 'dustmasker -in "%s" -out "%s" -outfmt fasta' % (
        args.fasta, outFilename)
    print('Running dustmasker. Hang in there... ', end=' ', file=sys.stderr)
    start = time()

    try:
        status = call(cmd, shell=True)
        print('done.\ndustmasker completed in %.2f seconds.' % (
            time() - start), file=sys.stderr)
        if status < 0:
            # Child terminated by signal.
            unlink(outFilename)
            sys.exit(-status)
        elif status > 0:
            print('WARNING: Child exited with status', status, file=sys.stderr)
            print('Output left in', outFilename, file=sys.stderr)
            sys.exit(status)
    except OSError as e:
        print('Execution failed:', e, file=sys.stderr)
        unlink(outFilename)
        sys.exit(1)

    save = sys.stdout.write
    reads = FastaReads(outFilename)
    maxLowComplexity = args.maxLowComplexity
    readCount = rejectCount = 0

    for read in reads:
        readCount += 1
        fasta = read.toString('fasta')
        if read.lowComplexityFraction() < maxLowComplexity:
            save(fasta)
        else:
            reject(fasta)
            rejectCount += 1

    unlink(outFilename)

    if args.rejectFile is not None:
        rejectFp.close()

    print((
        '%d sequences read, %d (%.2f%%) saved, %d (%.2f%%) rejected.' % (
            readCount, readCount - rejectCount,
            (readCount - rejectCount) / float(readCount) * 100.0,
            rejectCount,
            rejectCount / float(readCount) * 100.0)), file=sys.stderr)



import subprocess
from io import StringIO
from Bio import SeqIO


def getSequence(title, db='nt'):
    """
    @param title: A C{str} sequence title from a BLAST hit. Of the form
        'gi|63148399|gb|DQ011818.1| Description...'.
    @param db: the C{str} name of the BLAST database to search.
    @return: A C{SeqIO.read} instance.
    """
    titleId = title.split(' ', 1)[0]
    fasta = subprocess.check_output(
        ['blastdbcmd', '-entry', titleId, '-db', db]).decode('ascii')
    return SeqIO.read(StringIO(fasta), 'fasta')



#!/usr/bin/env python

"""
Read DNA FASTA from stdin and print FASTA to stdout, with the sequences
randomized. The read ids and lengths are preserved.

Note: This produces DNA sequences. If you have AA reads and you need this
      functionality, we can add it.
"""

from __future__ import print_function

import sys
from numpy.random import choice

from dark.fasta import FastaReads
from dark.reads import DNARead


if __name__ == '__main__':
    for read in FastaReads(sys.stdin):
        seq = ''.join(choice(['A', 'C', 'G', 'T'], len(read), replace=True))
        print(DNARead(read.id, seq).toString('fasta'))



#!/usr/bin/env python

"""
Read AA FASTA from stdin and print properties FASTA to stdout.
"""

from __future__ import print_function

import sys

from dark.fasta import FastaReads


if __name__ == '__main__':
    if len(sys.argv) > 1:
        print('Usage: %s < aaFile > propertiesFile.fasta' %
              sys.argv[0], file=sys.stderr)
        sys.exit(1)

    reads = FastaReads(sys.stdin)
    write = sys.stdout.write

    for read in reads:
        for aa in read.aaToProperties():
            write(aa.toString('fasta'))



import six
from unittest import TestCase

from dark.reads import Read
from dark.local_align import LocalAlignment


class TestLocalAlign(TestCase):
    """
    Test the LocalAlignment class.
    With match +1, mismatch -1, gap open -1, gap extend -1 and
        gap extend decay 0.0.
    """

    def testNonDNAString(self):
        """
        If the sequences do not consist of nucleotides, an exception
        must be raised.
        """
        self.skipTest('Non-DNA raising of ValueError is disabled.')
        seq1 = Read('seq1', 'xxx')
        seq2 = Read('seq2', 'yyy')
        six.assertRaisesRegex(self, ValueError, 'Invalid DNA nucleotide: "X"',
                              LocalAlignment, seq1, seq2)

    def testPositiveMismatch(self):
        """
        If the mismatch value passed is positive, an exception
        must be raised.
        """
        seq1 = Read('seq1', 'a')
        seq2 = Read('seq2', 'a')
        six.assertRaisesRegex(self, ValueError, 'Mismatch must be negative',
                              LocalAlignment, seq1, seq2, mismatch=3)

    def testZeroMismatch(self):
        """
        If the mismatch value passed is zero, an exception
        must be raised.
        """
        seq1 = Read('seq1', 'a')
        seq2 = Read('seq2', 'a')
        six.assertRaisesRegex(self, ValueError, 'Mismatch must be negative',
                              LocalAlignment, seq1, seq2, mismatch=0)

    def testPositiveGap(self):
        """
        If the gap value passed is positive, an exception
        must be raised.
        """
        seq1 = Read('seq1', 'a')
        seq2 = Read('seq2', 'a')
        six.assertRaisesRegex(self, ValueError, 'Gap must be negative',
                              LocalAlignment, seq1, seq2, gap=3)

    def testZeroGap(self):
        """
        If the gap value passed is zero, an exception
        must be raised.
        """
        seq1 = Read('seq1', 'a')
        seq2 = Read('seq2', 'a')
        six.assertRaisesRegex(self, ValueError, 'Gap must be negative',
                              LocalAlignment, seq1, seq2, gap=0)

    def testPositiveGapExtend(self):
        """
        If the gap extend value passed is positive, an exception
        must be raised.
        """
        seq1 = Read('seq1', 'a')
        seq2 = Read('seq2', 'a')
        six.assertRaisesRegex(self, ValueError,
                              'Gap extension penalty cannot be positive',
                              LocalAlignment, seq1, seq2, gapExtend=3)

    def testFirstSequenceEmpty(self):
        """
        If the first sequence passed is empty, an exception must be raised.
        """
        seq1 = Read('seq1', '')
        seq2 = Read('seq2', 'agtcagtcagtc')
        six.assertRaisesRegex(self, ValueError, 'Empty sequence: seq1',
                              LocalAlignment, seq1, seq2)

    def testSecondSequenceEmpty(self):
        """
        If the second sequence passed is empty, an exception must be raised.
        """
        seq1 = Read('seq1', 'agtcagtcagtc')
        seq2 = Read('seq2', '')
        six.assertRaisesRegex(self, ValueError, 'Empty sequence: seq2',
                              LocalAlignment, seq1, seq2)

    def testBothSequencesEmpty(self):
        """
        If two empty sequences are passed, an exception must be raised.
        """
        seq1 = Read('seq1', '')
        seq2 = Read('seq2', '')
        six.assertRaisesRegex(self, ValueError, 'Empty sequence: seq1',
                              LocalAlignment, seq1, seq2)

    def testGapAtStartOfSeq1(self):
        seq1 = Read('seq1', 'gaatcg')
        seq2 = Read('seq2', 'cgaatcg')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 6=\n'
                     'seq1 Match start: 1 Match end: 6\n'
                     'seq2 Match start: 2 Match end: 7\n'
                     'seq1 1 GAATCG 6\n'
                     '       ||||||\n'
                     'seq2 2 GAATCG 7')
        self.assertEqual(result, alignment)

    def testGapAtStartOfSeq2(self):
        seq1 = Read('seq1', 'cgaatcg')
        seq2 = Read('seq2', 'gaatcg')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 6=\n'
                     'seq1 Match start: 2 Match end: 7\n'
                     'seq2 Match start: 1 Match end: 6\n'
                     'seq1 2 GAATCG 7\n'
                     '       ||||||\n'
                     'seq2 1 GAATCG 6')
        self.assertEqual(result, alignment)

    def testGapAtEndOfSeq1(self):
        seq1 = Read('seq1', 'cgaatc')
        seq2 = Read('seq2', 'cgaatcg')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 6=\n'
                     'seq1 Match start: 1 Match end: 6\n'
                     'seq2 Match start: 1 Match end: 6\n'
                     'seq1 1 CGAATC 6\n'
                     '       ||||||\n'
                     'seq2 1 CGAATC 6')
        self.assertEqual(result, alignment)

    def testGapAtEndOfSeq2(self):
        seq1 = Read('seq1', 'cgaatcg')
        seq2 = Read('seq2', 'cgaatc')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 6=\n'
                     'seq1 Match start: 1 Match end: 6\n'
                     'seq2 Match start: 1 Match end: 6\n'
                     'seq1 1 CGAATC 6\n'
                     '       ||||||\n'
                     'seq2 1 CGAATC 6')
        self.assertEqual(result, alignment)

    def testGapAtBothEndsOfSeq1(self):
        seq1 = Read('seq1', 'gaatc')
        seq2 = Read('seq2', 'cgaatcg')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 5=\n'
                     'seq1 Match start: 1 Match end: 5\n'
                     'seq2 Match start: 2 Match end: 6\n'
                     'seq1 1 GAATC 5\n'
                     '       |||||\n'
                     'seq2 2 GAATC 6')
        self.assertEqual(result, alignment)

    def testGapAtBothEndsOfSeq2(self):
        seq1 = Read('seq1', 'cgaatcg')
        seq2 = Read('seq2', 'gaatc')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 5=\n'
                     'seq1 Match start: 2 Match end: 6\n'
                     'seq2 Match start: 1 Match end: 5\n'
                     'seq1 2 GAATC 6\n'
                     '       |||||\n'
                     'seq2 1 GAATC 5')
        self.assertEqual(result, alignment)

    def testAlignmentWithGapInMiddle(self):
        seq1 = Read('seq1', 'agtcagtcagtc')
        seq2 = Read('seq2', 'cgaatcg')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 2=1D1=\n'
                     'seq1 Match start: 7 Match end: 10\n'
                     'seq2 Match start: 5 Match end: 7\n'
                     'seq1 7 TCAG 10\n'
                     '       || |\n'
                     'seq2 5 TC-G 7')
        self.assertEqual(result, alignment)

    def testTwoEqualSequences(self):
        """
        When two identical sequences are given, the result should
        show that the sequences completely match.
        """
        seq1 = Read('seq1', 'cgaatcg')
        seq2 = Read('seq2', 'cgaatcg')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 7=\n'
                     'seq1 Match start: 1 Match end: 7\n'
                     'seq2 Match start: 1 Match end: 7\n'
                     'seq1 1 CGAATCG 7\n'
                     '       |||||||\n'
                     'seq2 1 CGAATCG 7')
        self.assertEqual(result, alignment)

    def testTwoCompletelyDifferentSequences(self):
        """
        When two completely different sequences are given, the result
        should be the two sequences with an empty alignment.
        """
        seq1 = Read('seq1', 'aaaaaa')
        seq2 = Read('seq2', 'gggggg')
        align = LocalAlignment(seq1, seq2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nNo alignment between seq1 and seq2\n')
        self.assertEqual(result, alignment)

    def testWikiAnswer(self):
        """
        Test the example given in Wikipedia:
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        """
        seq1 = Read('seq1', 'ACACACTA')
        seq2 = Read('seq2', 'AGCACACA')
        align = LocalAlignment(seq1, seq2, match=2)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 1=1I5=1D1=\n'
                     'seq1 Match start: 1 Match end: 8\n'
                     'seq2 Match start: 1 Match end: 8\n'
                     'seq1 1 A-CACACTA 8\n'
                     '       | ||||| |\n'
                     'seq2 1 AGCACAC-A 8')
        self.assertEqual(result, alignment)

    def testWikiAnswerWithMatchOne(self):
        """
        Test the example given in Wikipedia
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        Wikipedia uses a match score of two, here we use a score of one.
        """
        seq1 = Read('seq1', 'ACACACTA')
        seq2 = Read('seq2', 'AGCACACA')
        align = LocalAlignment(seq1, seq2, match=1)
        result = align.createAlignment(resultFormat=str)
        alignment = ('\nCigar string of aligned region: 5=1D1=\n'
                     'seq1 Match start: 2 Match end: 8\n'
                     'seq2 Match start: 3 Match end: 8\n'
                     'seq1 2 CACACTA 8\n'
                     '       ||||| |\n'
                     'seq2 3 CACAC-A 8')
        self.assertEqual(result, alignment)

    def testWikiAnswerAsDict(self):
        """
        Test the example given in Wikipedia:
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        with the return result being a dict.
        """
        seq1 = Read('seq1', 'ACACACTA')
        seq2 = Read('seq2', 'AGCACACA')
        align = LocalAlignment(seq1, seq2, match=2)
        result = align.createAlignment()
        self.assertEqual(
            {
                'cigar': '1=1I5=1D1=',
                'sequence1Start': 1,
                'sequence1End': 8,
                'sequence2Start': 1,
                'sequence2End': 8,
                'text': [
                    'seq1 1 A-CACACTA 8',
                    '       | ||||| |',
                    'seq2 1 AGCACAC-A 8',
                ]
            },
            result
        )

    def testWikiAnswerWithMatchOneAsDict(self):
        """
        Test the example given in Wikipedia
        http://en.wikipedia.org/wiki/Smith%E2%80%93Waterman_algorithm
        Wikipedia uses a match score of two, here we use a score of one.
        Get the result as a dict.
        """
        seq1 = Read('seq1', 'ACACACTA')
        seq2 = Read('seq2', 'AGCACACA')
        align = LocalAlignment(seq1, seq2, match=1)
        result = align.createAlignment()
        self.assertEqual(
            {
                'cigar': '5=1D1=',
                'sequence1Start': 2,
                'sequence1End': 8,
                'sequence2Start': 3,
                'sequence2End': 8,
                'text': [
                    'seq1 2 CACACTA 8',
                    '       ||||| |',
                    'seq2 3 CACAC-A 8',
                ]
            },
            result
        )



from dark.utils import countPrint
try:
    from itertools import zip_longest
except ImportError:
    # zip_longest does not exist in Python 2.7 itertools. We should be able
    # to get it via from six.moves import zip_longest according to
    # https://pythonhosted.org/six/index.html?highlight=zip_longest but
    # that doesn't work for me.
    from itertools import izip_longest as zip_longest

# From https://en.wikipedia.org/wiki/Amino_acid
#
# Alanine          Ala     A
# Arginine         Arg     R
# Asparagine       Asn     N
# Aspartic acid    Asp     D
# Cysteine         Cys     C
# Glutamic acid    Glu     E
# Glutamine        Gln     Q
# Glycine          Gly     G
# Histidine        His     H
# Isoleucine       Ile     I
# Leucine          Leu     L
# Lysine           Lys     K
# Methionine       Met     M
# Phenylalanine    Phe     F
# Proline          Pro     P
# Serine           Ser     S
# Threonine        Thr     T
# Tryptophan       Trp     W
# Tyrosine         Tyr     Y
# Valine           Val     V

NAMES = {
    'A': 'Alanine',
    'R': 'Arginine',
    'N': 'Asparagine',
    'D': 'Aspartic acid',
    'C': 'Cysteine',
    'E': 'Glutamic acid',
    'Q': 'Glutamine',
    'G': 'Glycine',
    'H': 'Histidine',
    'I': 'Isoleucine',
    'L': 'Leucine',
    'K': 'Lysine',
    'M': 'Methionine',
    'F': 'Phenylalanine',
    'P': 'Proline',
    'S': 'Serine',
    'T': 'Threonine',
    'V': 'Valine',
    'W': 'Tryptophan',
    'Y': 'Tyrosine',
}

AA_LETTERS = sorted(NAMES.keys())

NAMES_TO_ABBREV1 = dict((name, abbrev1) for abbrev1, name in NAMES.items())

ABBREV3 = {
    'A': 'Ala',
    'R': 'Arg',
    'N': 'Asn',
    'D': 'Asp',
    'C': 'Cys',
    'E': 'Glu',
    'Q': 'Gln',
    'G': 'Gly',
    'H': 'His',
    'I': 'Ile',
    'L': 'Leu',
    'K': 'Lys',
    'M': 'Met',
    'F': 'Phe',
    'P': 'Pro',
    'S': 'Ser',
    'T': 'Thr',
    'V': 'Val',
    'W': 'Trp',
    'Y': 'Tyr',
}

ABBREV3_TO_ABBREV1 = dict((abbrev3, abbrev1)
                          for abbrev1, abbrev3 in ABBREV3.items())

HYDROPHOBIC = 0x0001
HYDROPHILIC = 0x0002
AROMATIC = 0x0004
SULPHUR = 0x0008
ALIPHATIC = 0x0010
HYDROXYLIC = 0x0020
TINY = 0x0040
SMALL = 0x0080
ACIDIC = 0x0100
BASIC_POSITIVE = 0x0200
NEGATIVE = 0x0400
POLAR = 0x0800
NONE = 0x1000


ALL_PROPERTIES = (
    ACIDIC, ALIPHATIC, AROMATIC, BASIC_POSITIVE, HYDROPHILIC,
    HYDROPHOBIC, HYDROXYLIC, NEGATIVE, NONE, POLAR, SMALL, SULPHUR, TINY)

PROPERTY_NAMES = {
    ACIDIC: 'Acidic',
    ALIPHATIC: 'Aliphatic',
    AROMATIC: 'Aromatic',
    BASIC_POSITIVE: 'Basic positive',
    HYDROPHILIC: 'Hydrophilic',
    HYDROPHOBIC: 'Hydrophobic',
    HYDROXYLIC: 'Hydroxylic',
    NEGATIVE: 'Negative',
    NONE: '<NONE>',
    POLAR: 'Polar',
    SMALL: 'Small',
    SULPHUR: 'Sulphur',
    TINY: 'Tiny',
}

PROPERTIES = {
    'A': HYDROPHOBIC | SMALL | TINY,
    'C': HYDROPHOBIC | SMALL | TINY | SULPHUR,
    'D': HYDROPHILIC | SMALL | POLAR | NEGATIVE,
    'E': HYDROPHILIC | NEGATIVE | ACIDIC,
    'F': HYDROPHOBIC | AROMATIC,
    'G': HYDROPHILIC | SMALL | TINY,
    'H': HYDROPHOBIC | AROMATIC | POLAR | BASIC_POSITIVE,
    'I': ALIPHATIC | HYDROPHOBIC,
    'K': HYDROPHOBIC | BASIC_POSITIVE | POLAR,
    'L': ALIPHATIC | HYDROPHOBIC,
    'M': HYDROPHOBIC | SULPHUR,
    'N': HYDROPHILIC | SMALL | POLAR | ACIDIC,
    'P': HYDROPHILIC | SMALL,
    'Q': HYDROPHILIC | POLAR | ACIDIC,
    'R': HYDROPHILIC | POLAR | BASIC_POSITIVE,
    'S': HYDROPHILIC | SMALL | POLAR | HYDROXYLIC,
    'T': HYDROPHOBIC | SMALL | HYDROXYLIC,
    'V': ALIPHATIC | HYDROPHOBIC | SMALL,
    'W': HYDROPHOBIC | AROMATIC | POLAR,
    'Y': HYDROPHOBIC | AROMATIC | POLAR,
}


# A table with which codons translate to which amino acids.
# Based on https://en.wikipedia.org/wiki/DNA_codon_table
#
# Note that the trailing commas are necessary in the AAs that only have one
# codon. If you omit them, the parentheses will not create a tuple.

CODONS = {
    'A': ('GCA', 'GCC', 'GCG', 'GCT',),
    'C': ('TGC', 'TGT',),
    'D': ('GAC', 'GAT',),
    'E': ('GAA', 'GAG',),
    'F': ('TTC', 'TTT',),
    'G': ('GGA', 'GGC', 'GGG', 'GGT',),
    'H': ('CAC', 'CAT',),
    'I': ('ATA', 'ATC', 'ATT',),
    'K': ('AAA', 'AAG',),
    'L': ('CTA', 'CTC', 'CTG', 'CTT', 'TTA', 'TTG',),
    'M': ('ATG',),
    'N': ('AAC', 'AAT',),
    'P': ('CCA', 'CCC', 'CCG', 'CCT',),
    'Q': ('CAA', 'CAG',),
    'R': ('AGA', 'AGG', 'CGA', 'CGC', 'CGG', 'CGT',),
    'S': ('AGC', 'AGT', 'TCA', 'TCC', 'TCG', 'TCT',),
    'T': ('ACA', 'ACC', 'ACG', 'ACT',),
    'V': ('GTA', 'GTC', 'GTG', 'GTT',),
    'W': ('TGG',),
    'Y': ('TAC', 'TAT',),
}

START_CODON = 'ATG'
STOP_CODONS = ('TAA', 'TAG', 'TGA',)


"""
The dictionary below contains for each amino acid the value for
each property scaled from -1 to 1.
For documentation, check https://notebooks.antigenic-cartography.org/barbara/
pages/features/aa-properties.html
"""

PROPERTY_DETAILS = {
    'A': {
        'aliphaticity': 0.305785123967,
        'aromaticity': -0.550128534704,
        'composition': -1.0,
        'hydrogenation': 0.8973042362,
        'hydropathy': 0.4,
        'hydroxythiolation': -0.265160523187,
        'iep': -0.191489361702,
        'polar requirement': -0.463414634146,
        'polarity': -0.20987654321,
        'volume': -0.664670658683,
    },
    'C': {
        'aliphaticity': -0.00826446280992,
        'aromaticity': -0.740359897172,
        'composition': 1.0,
        'hydrogenation': 0.240051347882,
        'hydropathy': 0.555555555556,
        'hydroxythiolation': 0.785969084423,
        'iep': -0.424280350438,
        'polar requirement': -1.0,
        'polarity': -0.851851851852,
        'volume': -0.377245508982,
    },
    'D': {
        'aliphaticity': -0.818181818182,
        'aromaticity': -1.0,
        'composition': 0.00363636363636,
        'hydrogenation': -0.90243902439,
        'hydropathy': -0.777777777778,
        'hydroxythiolation': -0.348394768133,
        'iep': -1.0,
        'polar requirement': 1.0,
        'polarity': 1.0,
        'volume': -0.389221556886,
    },
    'E': {
        'aliphaticity': -0.553719008264,
        'aromaticity': -0.899742930591,
        'composition': -0.330909090909,
        'hydrogenation': -1.0,
        'hydropathy': -0.777777777778,
        'hydroxythiolation': -0.555291319857,
        'iep': -0.887359198999,
        'polar requirement': 0.878048780488,
        'polarity': 0.827160493827,
        'volume': -0.0419161676647,
    },
    'F': {
        'aliphaticity': 0.223140495868,
        'aromaticity': 0.858611825193,
        'composition': -1.0,
        'hydrogenation': 0.0218228498074,
        'hydropathy': 0.622222222222,
        'hydroxythiolation': 0.0582639714625,
        'iep': -0.321652065081,
        'polar requirement': -0.951219512195,
        'polarity': -0.925925925926,
        'volume': 0.544910179641,
    },
    'G': {
        'aliphaticity': -1.0,
        'aromaticity': -0.45501285347,
        'composition': -0.461818181818,
        'hydrogenation': 1.0,
        'hydropathy': -0.0888888888889,
        'hydroxythiolation': -0.158145065398,
        'iep': -0.198998748436,
        'polar requirement': -0.243902439024,
        'polarity': 0.0123456790123,
        'volume': -1.0,
    },
    'H': {
        'aliphaticity': -0.256198347107,
        'aromaticity': 0.555269922879,
        'composition': -0.578181818182,
        'hydrogenation': -0.150192554557,
        'hydropathy': -0.711111111111,
        'hydroxythiolation': 0.0154577883472,
        'iep': 0.206508135169,
        'polar requirement': -0.121951219512,
        'polarity': 0.358024691358,
        'volume': 0.11377245509,
    },
    'I': {
        'aliphaticity': 0.867768595041,
        'aromaticity': -0.264781491003,
        'composition': -1.0,
        'hydrogenation': 0.432605905006,
        'hydropathy': 1.0,
        'hydroxythiolation': -0.85255648038,
        'iep': -0.18648310388,
        'polar requirement': -0.975609756098,
        'polarity': -0.925925925926,
        'volume': 0.293413173653,
    },
    'K': {
        'aliphaticity': 0.123966942149,
        'aromaticity': -0.141388174807,
        'composition': -0.76,
        'hydrogenation': -0.142490372272,
        'hydropathy': -0.866666666667,
        'hydroxythiolation': -1.0,
        'iep': 0.744680851064,
        'polar requirement': 0.292682926829,
        'polarity': 0.58024691358,
        'volume': 0.389221556886,
    },
    'L': {
        'aliphaticity': 1.0,
        'aromaticity': -0.287917737789,
        'composition': -1.0,
        'hydrogenation': 0.381258023107,
        'hydropathy': 0.844444444444,
        'hydroxythiolation': -0.745541022592,
        'iep': -0.196495619524,
        'polar requirement': -0.975609756098,
        'polarity': -1.0,
        'volume': 0.293413173653,
    },
    'M': {
        'aliphaticity': 0.537190082645,
        'aromaticity': -0.372750642674,
        'composition': -1.0,
        'hydrogenation': -0.186136071887,
        'hydropathy': 0.422222222222,
        'hydroxythiolation': 0.0653983353151,
        'iep': -0.256570713392,
        'polar requirement': -0.878048780488,
        'polarity': -0.802469135802,
        'volume': 0.221556886228,
    },
    'N': {
        'aliphaticity': 0.471074380165,
        'aromaticity': -0.616966580977,
        'composition': -0.0327272727273,
        'hydrogenation': -0.548138639281,
        'hydropathy': -0.777777777778,
        'hydroxythiolation': 0.277051129608,
        'iep': -0.339173967459,
        'polar requirement': 0.268292682927,
        'polarity': 0.654320987654,
        'volume': -0.365269461078,
    },
    'P': {
        'aliphaticity': -0.917355371901,
        'aromaticity': -0.308483290488,
        'composition': -0.716363636364,
        'hydrogenation': 1.0,
        'hydropathy': -0.355555555556,
        'hydroxythiolation': -0.203329369798,
        'iep': -0.116395494368,
        'polar requirement': -0.560975609756,
        'polarity': -0.234567901235,
        'volume': -0.646706586826,
    },
    'Q': {
        'aliphaticity': 0.652892561983,
        'aromaticity': -0.439588688946,
        'composition': -0.352727272727,
        'hydrogenation': -0.602053915276,
        'hydropathy': -0.777777777778,
        'hydroxythiolation': -0.177170035672,
        'iep': -0.279098873592,
        'polar requirement': -0.0731707317073,
        'polarity': 0.382716049383,
        'volume': -0.0179640718563,
    },
    'R': {
        'aliphaticity': -0.157024793388,
        'aromaticity': -0.0642673521851,
        'composition': -0.527272727273,
        'hydrogenation': -0.401797175866,
        'hydropathy': -1.0,
        'hydroxythiolation': -0.51486325802,
        'iep': 1.0,
        'polar requirement': 0.0487804878049,
        'polarity': 0.382716049383,
        'volume': 0.449101796407,
    },
    'S': {
        'aliphaticity': 0.256198347107,
        'aromaticity': -0.660668380463,
        'composition': 0.0327272727273,
        'hydrogenation': 0.106546854942,
        'hydropathy': -0.177777777778,
        'hydroxythiolation': 1.0,
        'iep': -0.271589486859,
        'polar requirement': -0.341463414634,
        'polarity': 0.0617283950617,
        'volume': -0.652694610778,
    },
    'T': {
        'aliphaticity': -0.123966942149,
        'aromaticity': -0.80205655527,
        'composition': -0.483636363636,
        'hydrogenation': 0.399229781772,
        'hydropathy': -0.155555555556,
        'hydroxythiolation': 0.709869203329,
        'iep': -0.151439299124,
        'polar requirement': -0.560975609756,
        'polarity': -0.0864197530864,
        'volume': -0.305389221557,
    },
    'V': {
        'aliphaticity': 0.570247933884,
        'aromaticity': -0.665809768638,
        'composition': -1.0,
        'hydrogenation': 0.679075738126,
        'hydropathy': 0.933333333333,
        'hydroxythiolation': -0.621878715815,
        'iep': -0.201501877347,
        'polar requirement': -0.80487804878,
        'polarity': -0.753086419753,
        'volume': -0.0299401197605,
    },
    'W': {
        'aliphaticity': -0.619834710744,
        'aromaticity': 1.0,
        'composition': -0.905454545455,
        'hydrogenation': 0.0218228498074,
        'hydropathy': -0.2,
        'hydroxythiolation': 0.00118906064209,
        'iep': -0.219023779725,
        'polar requirement': -0.90243902439,
        'polarity': -0.876543209877,
        'volume': 1.0,
    },
    'Y': {
        'aliphaticity': -0.454545454545,
        'aromaticity': 0.712082262211,
        'composition': -0.854545454545,
        'hydrogenation': -0.304236200257,
        'hydropathy': 0.288888888889,
        'hydroxythiolation': 0.405469678954,
        'iep': -0.276595744681,
        'polar requirement': -0.853658536585,
        'polarity': -0.679012345679,
        'volume': 0.592814371257,
    },
}


"""
The dictionary below contains for each amino acid the value for
each property.
"""

PROPERTY_DETAILS_RAW = {
    'A': {
        'aliphaticity': 0.239,
        'aromaticity': -0.11,
        'composition': 0.0,
        'hydrogenation': 0.33,
        'hydropathy': 1.8,
        'hydroxythiolation': -0.062,
        'iep': 6.0,
        'polar requirement': 7.0,
        'polarity': 8.1,
        'volume': 31.0,
    },
    'C': {
        'aliphaticity': 0.22,
        'aromaticity': -0.184,
        'composition': 2.75,
        'hydrogenation': 0.074,
        'hydropathy': 2.5,
        'hydroxythiolation': 0.38,
        'iep': 5.07,
        'polar requirement': 4.8,
        'polarity': 5.5,
        'volume': 55.0,
    },
    'D': {
        'aliphaticity': 0.171,
        'aromaticity': -0.285,
        'composition': 1.38,
        'hydrogenation': -0.371,
        'hydropathy': -3.5,
        'hydroxythiolation': -0.079,
        'iep': 2.77,
        'polar requirement': 13.0,
        'polarity': 13.0,
        'volume': 54.0,
    },
    'E': {
        'aliphaticity': 0.187,
        'aromaticity': -0.246,
        'composition': 0.92,
        'hydrogenation': -0.409,
        'hydropathy': -3.5,
        'hydroxythiolation': -0.184,
        'iep': 3.22,
        'polar requirement': 12.5,
        'polarity': 12.3,
        'volume': 83.0,
    },
    'F': {
        'aliphaticity': 0.234,
        'aromaticity': 0.438,
        'composition': 0.0,
        'hydrogenation': -0.011,
        'hydropathy': 2.8,
        'hydroxythiolation': 0.074,
        'iep': 5.48,
        'polar requirement': 5.0,
        'polarity': 5.4,
        'volume': 132.0,
    },
    'G': {
        'aliphaticity': 0.16,
        'aromaticity': -0.073,
        'composition': 0.74,
        'hydrogenation': 0.37,
        'hydropathy': -0.4,
        'hydroxythiolation': -0.017,
        'iep': 5.97,
        'polar requirement': 7.9,
        'polarity': 9.0,
        'volume': 3.0,
    },
    'H': {
        'aliphaticity': 0.205,
        'aromaticity': 0.32,
        'composition': 0.58,
        'hydrogenation': -0.078,
        'hydropathy': -3.2,
        'hydroxythiolation': 0.056,
        'iep': 7.59,
        'polar requirement': 8.4,
        'polarity': 10.4,
        'volume': 96.0,
    },
    'I': {
        'aliphaticity': 0.273,
        'aromaticity': 0.001,
        'composition': 0.0,
        'hydrogenation': 0.149,
        'hydropathy': 4.5,
        'hydroxythiolation': -0.309,
        'iep': 6.02,
        'polar requirement': 4.9,
        'polarity': 5.2,
        'volume': 111.0,
    },
    'K': {
        'aliphaticity': 0.228,
        'aromaticity': 0.049,
        'composition': 0.33,
        'hydrogenation': -0.075,
        'hydropathy': -3.9,
        'hydroxythiolation': -0.371,
        'iep': 9.74,
        'polar requirement': 10.1,
        'polarity': 11.3,
        'volume': 119.0,
    },
    'L': {
        'aliphaticity': 0.281,
        'aromaticity': -0.008,
        'composition': 0.0,
        'hydrogenation': 0.129,
        'hydropathy': 3.8,
        'hydroxythiolation': -0.264,
        'iep': 5.98,
        'polar requirement': 4.9,
        'polarity': 4.9,
        'volume': 111.0,
    },
    'M': {
        'aliphaticity': 0.253,
        'aromaticity': -0.041,
        'composition': 0.0,
        'hydrogenation': -0.092,
        'hydropathy': 1.9,
        'hydroxythiolation': 0.077,
        'iep': 5.74,
        'polar requirement': 5.3,
        'polarity': 5.7,
        'volume': 105.0,
    },
    'N': {
        'aliphaticity': 0.249,
        'aromaticity': -0.136,
        'composition': 1.33,
        'hydrogenation': -0.233,
        'hydropathy': -3.5,
        'hydroxythiolation': 0.166,
        'iep': 5.41,
        'polar requirement': 10.0,
        'polarity': 11.6,
        'volume': 56.0,
    },
    'P': {
        'aliphaticity': 0.165,
        'aromaticity': -0.016,
        'composition': 0.39,
        'hydrogenation': 0.37,
        'hydropathy': -1.6,
        'hydroxythiolation': -0.036,
        'iep': 6.3,
        'polar requirement': 6.6,
        'polarity': 8.0,
        'volume': 32.5,
    },
    'Q': {
        'aliphaticity': 0.26,
        'aromaticity': -0.067,
        'composition': 0.89,
        'hydrogenation': -0.254,
        'hydropathy': -3.5,
        'hydroxythiolation': -0.025,
        'iep': 5.65,
        'polar requirement': 8.6,
        'polarity': 10.5,
        'volume': 85.0,
    },
    'R': {
        'aliphaticity': 0.211,
        'aromaticity': 0.079,
        'composition': 0.65,
        'hydrogenation': -0.176,
        'hydropathy': -4.5,
        'hydroxythiolation': -0.167,
        'iep': 10.76,
        'polar requirement': 9.1,
        'polarity': 10.5,
        'volume': 124.0,
    },
    'S': {
        'aliphaticity': 0.236,
        'aromaticity': -0.153,
        'composition': 1.42,
        'hydrogenation': 0.022,
        'hydropathy': -0.8,
        'hydroxythiolation': 0.47,
        'iep': 5.68,
        'polar requirement': 7.5,
        'polarity': 9.2,
        'volume': 32.0,
    },
    'T': {
        'aliphaticity': 0.213,
        'aromaticity': -0.208,
        'composition': 0.71,
        'hydrogenation': 0.136,
        'hydropathy': -1.3,
        'hydroxythiolation': 0.348,
        'iep': 6.16,
        'polar requirement': 6.6,
        'polarity': 8.6,
        'volume': 61.0,
    },
    'V': {
        'aliphaticity': 0.255,
        'aromaticity': -0.155,
        'composition': 0.0,
        'hydrogenation': 0.245,
        'hydropathy': 4.2,
        'hydroxythiolation': -0.212,
        'iep': 5.96,
        'polar requirement': 5.6,
        'polarity': 5.9,
        'volume': 84.0,
    },
    'W': {
        'aliphaticity': 0.183,
        'aromaticity': 0.493,
        'composition': 0.13,
        'hydrogenation': -0.011,
        'hydropathy': -0.9,
        'hydroxythiolation': 0.05,
        'iep': 5.89,
        'polar requirement': 5.2,
        'polarity': 5.4,
        'volume': 170.0,
    },
    'Y': {
        'aliphaticity': 0.193,
        'aromaticity': 0.183,
        'composition': 0.2,
        'hydrogenation': -0.138,
        'hydropathy': -1.3,
        'hydroxythiolation': 0.22,
        'iep': 5.66,
        'polar requirement': 5.4,
        'polarity': 6.2,
        'volume': 136.0,
    },
}


"""
Clusters based on raw amino acid property values. See
https://notebooks.antigenic-cartography.org/barbara/pages/features/
aa-properties.html and https://notebooks.antigenic-cartography.org/barbara/
pages/features/new-tps.html
"""

PROPERTY_CLUSTERS = {
    'A': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 3,
        'hydroxythiolation': 2,
        'iep': 2,
        'polar requirement': 2,
        'polarity': 2,
        'volume': 2,
    },
    'C': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 3,
        'hydrogenation': 1,
        'hydropathy': 3,
        'hydroxythiolation': 5,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 3,
    },
    'D': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 2,
        'hydrogenation': 1,
        'hydropathy': 1,
        'hydroxythiolation': 2,
        'iep': 1,
        'polar requirement': 4,
        'polarity': 4,
        'volume': 3,
    },
    'E': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 1,
        'hydroxythiolation': 1,
        'iep': 1,
        'polar requirement': 4,
        'polarity': 4,
        'volume': 4,
    },
    'F': {
        'aliphaticity': 1,
        'aromaticity': 2,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 3,
        'hydroxythiolation': 3,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 4,
    },
    'G': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 2,
        'hydroxythiolation': 2,
        'iep': 2,
        'polar requirement': 2,
        'polarity': 2,
        'volume': 1,
    },
    'H': {
        'aliphaticity': 1,
        'aromaticity': 2,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 1,
        'hydroxythiolation': 3,
        'iep': 3,
        'polar requirement': 2,
        'polarity': 3,
        'volume': 4,
    },
    'I': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 4,
        'hydroxythiolation': 1,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 4,
    },
    'K': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 1,
        'hydroxythiolation': 1,
        'iep': 3,
        'polar requirement': 3,
        'polarity': 4,
        'volume': 4,
    },
    'L': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 4,
        'hydroxythiolation': 1,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 4,
    },
    'M': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 3,
        'hydroxythiolation': 3,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 4,
    },
    'N': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 2,
        'hydrogenation': 1,
        'hydropathy': 1,
        'hydroxythiolation': 4,
        'iep': 2,
        'polar requirement': 3,
        'polarity': 4,
        'volume': 3,
    },
    'P': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 2,
        'hydroxythiolation': 2,
        'iep': 2,
        'polar requirement': 2,
        'polarity': 2,
        'volume': 2,
    },
    'Q': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 1,
        'hydroxythiolation': 2,
        'iep': 2,
        'polar requirement': 2,
        'polarity': 3,
        'volume': 4,
    },
    'R': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 1,
        'hydroxythiolation': 1,
        'iep': 3,
        'polar requirement': 2,
        'polarity': 3,
        'volume': 4,
    },
    'S': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 2,
        'hydrogenation': 1,
        'hydropathy': 2,
        'hydroxythiolation': 5,
        'iep': 2,
        'polar requirement': 2,
        'polarity': 2,
        'volume': 2,
    },
    'T': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 2,
        'hydroxythiolation': 5,
        'iep': 2,
        'polar requirement': 2,
        'polarity': 2,
        'volume': 3,
    },
    'V': {
        'aliphaticity': 1,
        'aromaticity': 1,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 4,
        'hydroxythiolation': 1,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 4,
    },
    'W': {
        'aliphaticity': 1,
        'aromaticity': 2,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 2,
        'hydroxythiolation': 3,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 5,
    },
    'Y': {
        'aliphaticity': 1,
        'aromaticity': 2,
        'composition': 1,
        'hydrogenation': 1,
        'hydropathy': 2,
        'hydroxythiolation': 4,
        'iep': 2,
        'polar requirement': 1,
        'polarity': 1,
        'volume': 4,
    },
}


class AminoAcid(object):
    """
    Hold information about an amino acid.

    @param name: The full C{str} name of the amino acid.
    @param abbrev3: The 3-letter C{str} abbreviation of the amino acid,
        e.g., 'Arg'.
    @param abbrev1: The 1-letter C{str} abbreviation of the amino acid,
        e.g., 'A'.
    @param codons: A C{list} of 3-letter codons for the amino acid.
    @param properties: An C{int} logical-AND of the various properties
        (see PROPERTIES, above) of this amino acid.
    @param propertyDetails: A C{dict} containing property names and values
        for this amino acid. E.g.:
        {
            'aliphaticity': -0.157024793388,
            'aromaticity': -0.0642673521851,
            'composition': -0.527272727273,
            'hydrogenation': -0.401797175866,
            'hydropathy': -1.0,
            'hydroxythiolation': -0.51486325802,
            'iep': 1.0,
            'polar requirement': 0.0487804878049,
            'polarity': 0.382716049383,
            'volume': 0.449101796407,
        }
    @param propertyClusters: A C{dict} containing the property names and
        clusters for this amino acid. E.g.:
        {
            'aliphaticity': 1,
            'aromaticity': 2,
            'composition': 1,
            'hydrogenation': 1,
            'hydropathy': 2,
            'hydroxythiolation': 4,
            'iep': 2,
            'polar requirement': 1,
            'polarity': 1,
            'volume': 4,
        }
    """
    def __init__(self, name, abbrev3, abbrev1, codons, properties,
                 propertyDetails, propertyClusters):
        self.name = name
        self.abbrev3 = abbrev3
        self.abbrev1 = abbrev1
        self.codons = codons
        self.properties = properties
        self.propertyDetails = propertyDetails
        self.propertyClusters = propertyClusters


def find(s):
    """
    Find an amino acid whose name or abbreviation is s.

    @param s: A C{str} amino acid specifier. This may be a full name,
        a 3-letter abbreviation or a 1-letter abbreviation. Case is ignored.
    return: An L{AminoAcid} instance or C{None} if no matching amino acid can
        be located.
    """

    abbrev1 = None
    origS = s

    if ' ' in s:
        # Convert first word to title case, others to lower.
        first, rest = s.split(' ', 1)
        s = first.title() + ' ' + rest.lower()
    else:
        s = s.title()

    if s in NAMES:
        abbrev1 = s
    elif s in ABBREV3_TO_ABBREV1:
        abbrev1 = ABBREV3_TO_ABBREV1[s]
    elif s in NAMES_TO_ABBREV1:
        abbrev1 = NAMES_TO_ABBREV1[s]
    else:
        # Look for a 3-letter codon.
        def findCodon(target):
            for abbrev1, codons in CODONS.items():
                for codon in codons:
                    if codon == target:
                        return abbrev1

        abbrev1 = findCodon(origS.upper())

    if abbrev1:
        return AminoAcid(
            NAMES[abbrev1], ABBREV3[abbrev1], abbrev1, CODONS[abbrev1],
            PROPERTIES[abbrev1], PROPERTY_DETAILS[abbrev1],
            PROPERTY_CLUSTERS[abbrev1])


def _propertiesOrClustersForSequence(sequence, propertyNames, propertyValues,
                                     missingAAValue):
    """
    Extract amino acid property values or cluster numbers for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{propertyValues} C{dict}).
    @param propertyValues: A C{dict} in the form of C{PROPERTY_DETAILS} or
        C{PROPERTY_CLUSTERS} (see above).
    @param missingAAValue: A C{float} value to use for properties when an AA
        (e.g., 'X') is not known.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: A C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding property value in C{propertyValues} in
        order of sequence position.
    """
    propertyNames = sorted(map(str.lower, set(propertyNames)))

    # Make sure all mentioned property names exist for at least one AA.
    knownProperties = set()
    for names in propertyValues.values():
        knownProperties.update(names)
    unknown = set(propertyNames) - knownProperties
    if unknown:
        raise ValueError(
            'Unknown propert%s: %s.' %
            ('y' if len(unknown) == 1 else 'ies', ', '.join(unknown)))

    aas = sequence.sequence.upper()
    result = {}

    for propertyName in propertyNames:
        result[propertyName] = values = []
        append = values.append
        for aa in aas:
            try:
                properties = propertyValues[aa]
            except KeyError:
                # No such AA.
                append(missingAAValue)
            else:
                append(properties[propertyName])

    return result


def propertiesForSequence(sequence, propertyNames, missingAAValue=-1.1):
    """
    Extract amino acid property values for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{dark.aa.PROPERTY_DETAILS} C{dict}).
    @param missingAAValue: A C{float} value to use for properties when an AA
        (e.g., 'X') is not known.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: A C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding property value according to sequence
        position.
    """
    return _propertiesOrClustersForSequence(
        sequence, propertyNames, PROPERTY_DETAILS, missingAAValue)


def clustersForSequence(sequence, propertyNames, missingAAValue=0):
    """
    Extract amino acid property cluster numbers for a sequence.

    @param sequence: An C{AARead} (or a subclass) instance.
    @param propertyNames: An iterable of C{str} property names (each of which
        must be a key of a key in the C{dark.aa.PROPERTY_CLUSTERS} C{dict}).
    @param missingAAValue: An C{int} value to use for properties when an AA
        (e.g., 'X') is not known.
    @raise ValueError: If an unknown property is given in C{propertyNames}.
    @return: A C{dict} keyed by (lowercase) property name, with values that are
        C{list}s of the corresponding property cluster number according to
        sequence position.
    """
    return _propertiesOrClustersForSequence(
        sequence, propertyNames, PROPERTY_CLUSTERS, missingAAValue)


def matchToString(aaMatch, read1, read2, indent='', offsets=None):
    """
    Format amino acid sequence match as a string.

    @param aaMatch: A C{dict} returned by C{compareAaReads}.
    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param indent: A C{str} to indent all returned lines with.
    @param offsets: If not C{None}, a C{set} of offsets of interest that were
        only considered when making C{match}.
    @return: A C{str} describing the match.
    """
    match = aaMatch['match']
    matchCount = match['matchCount']
    gapMismatchCount = match['gapMismatchCount']
    gapGapMismatchCount = match['gapGapMismatchCount']
    nonGapMismatchCount = match['nonGapMismatchCount']

    if offsets:
        len1 = len2 = len(offsets)
    else:
        len1, len2 = map(len, (read1, read2))

    result = []
    append = result.append

    append(countPrint('%sMatches' % indent, matchCount, len1, len2))
    mismatchCount = (gapMismatchCount + gapGapMismatchCount +
                     nonGapMismatchCount)
    append(countPrint('%sMismatches' % indent, mismatchCount, len1, len2))
    append(countPrint('%s  Not involving gaps (i.e., conflicts)' % (indent),
                      nonGapMismatchCount, len1, len2))
    append(countPrint('%s  Involving a gap in one sequence' % indent,
                      gapMismatchCount, len1, len2))
    append(countPrint('%s  Involving a gap in both sequences' % indent,
                      gapGapMismatchCount, len1, len2))

    for read, key in zip((read1, read2), ('read1', 'read2')):
        append('%s  Id: %s' % (indent, read.id))
        length = len(read)
        append('%s    Length: %d' % (indent, length))
        gapCount = len(aaMatch[key]['gapOffsets'])
        append(countPrint('%s    Gaps' % indent, gapCount, length))
        if gapCount:
            append(
                '%s    Gap locations (1-based): %s' %
                (indent,
                 ', '.join(map(lambda offset: str(offset + 1),
                               sorted(aaMatch[key]['gapOffsets'])))))
        extraCount = aaMatch[key]['extraCount']
        if extraCount:
            append(countPrint('%s    Extra nucleotides at end' % indent,
                              extraCount, length))

    return '\n'.join(result)


def compareAaReads(read1, read2, gapChars='-', offsets=None):
    """
    Compare two amino acid sequences.

    @param read1: A C{Read} instance or an instance of one of its subclasses.
    @param read2: A C{Read} instance or an instance of one of its subclasses.
    @param gapChars: An object supporting __contains__ with characters that
        should be considered to be gaps.
    @param offsets: If not C{None}, a C{set} of offsets of interest. Offsets
        not in the set will not be considered.
    @return: A C{dict} with information about the match and the individual
        sequences (see below).
    """
    matchCount = 0
    gapMismatchCount = nonGapMismatchCount = gapGapMismatchCount = 0
    read1ExtraCount = read2ExtraCount = 0
    read1GapOffsets = []
    read2GapOffsets = []

    for offset, (a, b) in enumerate(zip_longest(read1.sequence.upper(),
                                                read2.sequence.upper())):
        # Use 'is not None' in the following to allow an empty offsets set
        # to be passed.
        if offsets is not None and offset not in offsets:
            continue
        if a is None:
            # b has an extra character at its end (it cannot be None).
            assert b is not None
            read2ExtraCount += 1
            if b in gapChars:
                read2GapOffsets.append(offset)
        elif b is None:
            # a has an extra character at its end.
            read1ExtraCount += 1
            if a in gapChars:
                read1GapOffsets.append(offset)
        else:
            # We have a character from both sequences (they could still be
            # gap characters).
            if a in gapChars:
                read1GapOffsets.append(offset)
                if b in gapChars:
                    # Both are gaps. This can happen (though hopefully not
                    # if the sequences were pairwise aligned).
                    gapGapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # a is a gap, b is not.
                    gapMismatchCount += 1
            else:
                if b in gapChars:
                    # b is a gap, a is not.
                    gapMismatchCount += 1
                    read2GapOffsets.append(offset)
                else:
                    # Neither is a gap character.
                    if a == b:
                        matchCount += 1
                    else:
                        nonGapMismatchCount += 1

    return {
        'match': {
            'matchCount': matchCount,
            'gapMismatchCount': gapMismatchCount,
            'gapGapMismatchCount': gapGapMismatchCount,
            'nonGapMismatchCount': nonGapMismatchCount,
        },
        'read1': {
            'extraCount': read1ExtraCount,
            'gapOffsets': read1GapOffsets,
        },
        'read2': {
            'extraCount': read2ExtraCount,
            'gapOffsets': read2GapOffsets,
        },
    }



#!/usr/bin/env python

import argparse
import bz2file
import sys

from dark.blast.conversion import XMLRecordsReader


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Convert a BLAST XML file to JSON.',
        epilog=('Give a BLAST XML file and convert it to JSON. Optionally '
                'compress the JSON output.')
    )

    parser.add_argument(
        '--json', metavar='JSON-output-file',
        help=('the JSON filename to write the converted XML to. If omitted, '
              'standard output will be used.'))

    parser.add_argument(
        '--xml', metavar='BLAST-XML-file', default=sys.stdin,
        help=('the BLAST XML output file to convert. If omitted, standard '
              'input will be read.'))

    parser.add_argument(
        '--bzip2', default=False, action='store_true',
        help='If True, compress the JSON output using bzip2.')

    args = parser.parse_args()

    if args.bzip2:
        fp = bz2file.BZ2File(args.json or sys.stdout, 'w')
    else:
        fp = open(args.json, 'w') if args.json else sys.stdout

    reader = XMLRecordsReader(args.xml)
    reader.saveAsJSON(fp)
    fp.close()



import six
from unittest import TestCase
from itertools import product

from dark.aa import (
    PROPERTIES, ALL_PROPERTIES, PROPERTY_NAMES, ACIDIC,
    ALIPHATIC, AROMATIC, BASIC_POSITIVE, HYDROPHILIC, HYDROPHOBIC,
    HYDROXYLIC, NEGATIVE, NONE, POLAR, SMALL, SULPHUR, TINY, NAMES,
    NAMES_TO_ABBREV1, ABBREV3, ABBREV3_TO_ABBREV1, CODONS, AA_LETTERS,
    find, AminoAcid, clustersForSequence, propertiesForSequence,
    PROPERTY_CLUSTERS, PROPERTY_DETAILS_RAW, START_CODON, STOP_CODONS,
    compareAaReads, matchToString)
from dark.reads import AARead


class TestAALetters(TestCase):
    """
    Test the AA_LETTERS string in dark.aa
    """

    def testExpectedAALetters(self):
        """
        The AA_LETTERS value must be as expected.
        """
        # From https://en.wikipedia.org/wiki/Amino_acid
        expected = sorted('ARNDCEQGHILKMFPSTWYV')
        self.assertEqual(expected, AA_LETTERS)


class TestProperties(TestCase):
    """
    Test the AA properties in dark.aa
    """

    def testCorrectAAs(self):
        """
        The PROPERTIES dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(PROPERTIES.keys()))

    def testAllProperties(self):
        """
        The ALL_PROPERTIES tuple must contain all known properties.
        """
        expected = set([
            ACIDIC, ALIPHATIC, AROMATIC, BASIC_POSITIVE, HYDROPHILIC,
            HYDROPHOBIC, HYDROXYLIC, NEGATIVE, NONE, POLAR, SMALL, SULPHUR,
            TINY])
        self.assertEqual(set(ALL_PROPERTIES), expected)

    def testPropertyValuesDiffer(self):
        """
        All individual property values must be different.
        """
        self.assertEqual(len(ALL_PROPERTIES), len(set(ALL_PROPERTIES)))

    def testEachAAHasItsCorrectPropertiesAndNoOthers(self):
        """
        Each amino acid must have the properties expected of it, and no others.
        """
        expected = {
            'A': set([HYDROPHOBIC, SMALL, TINY]),
            'C': set([HYDROPHOBIC, SMALL, TINY, SULPHUR]),
            'D': set([HYDROPHILIC, SMALL, POLAR, NEGATIVE]),
            'E': set([HYDROPHILIC, NEGATIVE, ACIDIC]),
            'F': set([HYDROPHOBIC, AROMATIC]),
            'G': set([HYDROPHILIC, SMALL, TINY]),
            'H': set([HYDROPHOBIC, AROMATIC, POLAR, BASIC_POSITIVE]),
            'I': set([ALIPHATIC, HYDROPHOBIC]),
            'K': set([HYDROPHOBIC, BASIC_POSITIVE, POLAR]),
            'L': set([ALIPHATIC, HYDROPHOBIC]),
            'M': set([HYDROPHOBIC, SULPHUR]),
            'N': set([HYDROPHILIC, SMALL, POLAR, ACIDIC]),
            'P': set([HYDROPHILIC, SMALL]),
            'Q': set([HYDROPHILIC, POLAR, ACIDIC]),
            'R': set([HYDROPHILIC, POLAR, BASIC_POSITIVE]),
            'S': set([HYDROPHILIC, SMALL, POLAR, HYDROXYLIC]),
            'T': set([HYDROPHOBIC, SMALL, HYDROXYLIC]),
            'V': set([ALIPHATIC, HYDROPHOBIC, SMALL]),
            'W': set([HYDROPHOBIC, AROMATIC, POLAR]),
            'Y': set([HYDROPHOBIC, AROMATIC, POLAR]),
        }

        # Make sure our 'expected' dict (above) has properties for everything.
        self.assertEqual(AA_LETTERS, sorted(expected.keys()))

        for aa in AA_LETTERS:
            for prop in ALL_PROPERTIES:
                if prop in expected[aa]:
                    self.assertTrue(bool(PROPERTIES[aa] & prop))
                else:
                    self.assertFalse(bool(PROPERTIES[aa] & prop))

    def testProperyNamesKeys(self):
        """
        The PROPERTY_NAMES dict must have a key for each property.
        """
        self.assertEqual(sorted(ALL_PROPERTIES), sorted(PROPERTY_NAMES.keys()))

    def testProperyNamesValuesDiffer(self):
        """
        The PROPERTY_NAMES dict must have different values for each key.
        """
        self.assertEqual(len(PROPERTY_NAMES),
                         len(set(PROPERTY_NAMES.values())))


class TestNames(TestCase):
    """
    Test the NAMES dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The NAMES dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(NAMES.keys()))

    def testValuesDiffer(self):
        """
        All individual names must be different.
        """
        names = list(NAMES.values())
        self.assertEqual(len(names), len(set(names)))


class TestNamesToAbbrev1(TestCase):
    """
    Test the NAMES_TO_ABBREV1 dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The NAMES_TO_ABBREV1 dict must have the correct AA values.
        """
        self.assertEqual(AA_LETTERS,
                         sorted(NAMES_TO_ABBREV1.values()))


class TestAbbrev3(TestCase):
    """
    Test the ABBREV3 dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The ABBREV3 dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(ABBREV3.keys()))

    def testValuesDiffer(self):
        """
        All individual names must be different.
        """
        names = list(ABBREV3.values())
        self.assertEqual(len(names), len(set(names)))


class TestAbbrev3ToAbbrev1(TestCase):
    """
    Test the ABBREV3_TO_ABBREV1 dict in dark.aa
    """

    def testCorrectAAs(self):
        """
        The ABBREV3_TO_ABBREV1 dict must have the correct AA values.
        """
        self.assertEqual(AA_LETTERS,
                         sorted(ABBREV3_TO_ABBREV1.values()))


class TestCodons(TestCase):
    """
    Tests for the CODONS dict.
    """

    def testCorrectAAs(self):
        """
        The CODONS dict must have the correct AA keys.
        """
        self.assertEqual(AA_LETTERS, sorted(CODONS.keys()))

    def testNumberCodons(self):
        """
        The table must contain 61 codons. This is 4^3 - 3. I.e., all possible
        combinations of 3 bases minus the three stop codons.
        """
        self.assertEqual(61, sum(len(codons) for codons in CODONS.values()))

    def testCodonLength(self):
        """
        All codons must be three bases long.
        """
        for codons in CODONS.values():
            self.assertTrue(all(len(codon) == 3 for codon in codons))

    def testCodonContent(self):
        """
        Codons must only contain the letters A, T, G, C.
        """
        for codons in CODONS.values():
            for codon in codons:
                self.assertTrue(all(letter in 'ACGT' for letter in codon))

    def testCodonsAreNotAbbrev3s(self):
        """
        No codon can be the same as an amino acid 3-letter abbreviation (or
        else our find function may not be unambiguous in what it returns).
        """
        for codons in CODONS.values():
            self.assertFalse(
                any(codon.title() in ABBREV3_TO_ABBREV1 for codon in codons))

    def testDistinct(self):
        """
        All nucleotide triples must be distinct.
        """
        seen = set()
        for codons in CODONS.values():
            for codon in codons:
                seen.add(codon)
        self.assertEqual(61, len(seen))

    def testStartCodon(self):
        """
        The start codon must have the expected value.
        """
        self.assertEqual('ATG', START_CODON)

    def testStartCodonIsMethionine(self):
        """
        The start codon is the same as Methionine.
        """
        self.assertEqual((START_CODON,), CODONS['M'])

    def testStopCodons(self):
        """
        The stop codons must have the expected value.
        """
        self.assertEqual(('TAA', 'TAG', 'TGA'), STOP_CODONS)

    def testStopCodonsNotInCodonTable(self):
        """
        The stop codons must not be in the main table.
        """
        for stop in STOP_CODONS:
            for codons in CODONS.values():
                self.assertNotIn(stop, codons)

    def testAllCodonsPresent(self):
        """
        All possible codons must be present, as either coding for an AA
        or as a stop codon.
        """
        combinations = set(
            ''.join(x) for x in product('ACGT', 'ACGT', 'ACGT'))
        for codons in CODONS.values():
            for codon in codons:
                combinations.remove(codon)

        # Just the stop codons should be left.
        self.assertEqual(set(STOP_CODONS), combinations)


class TestAminoAcid(TestCase):
    """
    Test the AminoAcid class in dark.aa
    """

    def testExpectedAttributes(self):
        """
        An amino acid instance must have the expected attributes.
        """
        properties = {}
        propertyDetails = {}
        codons = []
        propertyClusters = {}
        aa = AminoAcid('Alanine', 'Ala', 'A', codons, properties,
                       propertyDetails, propertyClusters)
        self.assertEqual('Alanine', aa.name)
        self.assertEqual('Ala', aa.abbrev3)
        self.assertEqual('A', aa.abbrev1)
        self.assertIs(codons, aa.codons)
        self.assertIs(properties, aa.properties)
        self.assertIs(propertyDetails, aa.propertyDetails)
        self.assertIs(propertyClusters, aa.propertyClusters)


class TestFind(TestCase):
    """
    Test the find function in dark.aa
    """

    def testFindUnknown(self):
        """
        find must return C{None} when called with an unrecognized value.
        """
        self.assertEqual(None, find('silly'))

    def testFindByAbbrev1(self):
        """
        It must be possible to find an amino acid by its 1-letter abbreviation.
        """
        aa = find('A')
        self.assertEqual('Alanine', aa.name)
        self.assertEqual('Ala', aa.abbrev3)
        self.assertEqual('A', aa.abbrev1)
        self.assertEqual(('GCA', 'GCC', 'GCG', 'GCT'), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                'aliphaticity': 0.305785123967,
                'aromaticity': -0.550128534704,
                'composition': -1.0,
                'hydrogenation': 0.8973042362,
                'hydropathy': 0.4,
                'hydroxythiolation': -0.265160523187,
                'iep': -0.191489361702,
                'polar requirement': -0.463414634146,
                'polarity': -0.20987654321,
                'volume': -0.664670658683,
            },
            aa.propertyDetails)
        self.assertEqual(
            {
                'aliphaticity': 1,
                'aromaticity': 1,
                'composition': 1,
                'hydrogenation': 1,
                'hydropathy': 3,
                'hydroxythiolation': 2,
                'iep': 2,
                'polar requirement': 2,
                'polarity': 2,
                'volume': 2,
            },
            aa.propertyClusters)

    def testFindByAbbrev3(self):
        """
        It must be possible to find an amino acid by its 3-letter abbreviation.
        """
        aa = find('Ala')
        self.assertEqual('Alanine', aa.name)
        self.assertEqual('Ala', aa.abbrev3)
        self.assertEqual('A', aa.abbrev1)
        self.assertEqual(('GCA', 'GCC', 'GCG', 'GCT'), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                'aliphaticity': 0.305785123967,
                'aromaticity': -0.550128534704,
                'composition': -1.0,
                'hydrogenation': 0.8973042362,
                'hydropathy': 0.4,
                'hydroxythiolation': -0.265160523187,
                'iep': -0.191489361702,
                'polar requirement': -0.463414634146,
                'polarity': -0.20987654321,
                'volume': -0.664670658683,
            },
            aa.propertyDetails)
        self.assertEqual(
            {
                'aliphaticity': 1,
                'aromaticity': 1,
                'composition': 1,
                'hydrogenation': 1,
                'hydropathy': 3,
                'hydroxythiolation': 2,
                'iep': 2,
                'polar requirement': 2,
                'polarity': 2,
                'volume': 2,
            },
            aa.propertyClusters)

    def testFindByCodon(self):
        """
        It must be possible to find an amino acid by a 3-letter codon.
        """
        aa = find('GCC')
        self.assertEqual('Alanine', aa.name)
        self.assertEqual('Ala', aa.abbrev3)
        self.assertEqual('A', aa.abbrev1)
        self.assertEqual(('GCA', 'GCC', 'GCG', 'GCT'), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                'aliphaticity': 0.305785123967,
                'aromaticity': -0.550128534704,
                'composition': -1.0,
                'hydrogenation': 0.8973042362,
                'hydropathy': 0.4,
                'hydroxythiolation': -0.265160523187,
                'iep': -0.191489361702,
                'polar requirement': -0.463414634146,
                'polarity': -0.20987654321,
                'volume': -0.664670658683,
            },
            aa.propertyDetails)
        self.assertEqual(
            {
                'aliphaticity': 1,
                'aromaticity': 1,
                'composition': 1,
                'hydrogenation': 1,
                'hydropathy': 3,
                'hydroxythiolation': 2,
                'iep': 2,
                'polar requirement': 2,
                'polarity': 2,
                'volume': 2,
            },
            aa.propertyClusters)

    def testFindByName(self):
        """
        It must be possible to find an amino acid by its name.
        """
        aa = find('Alanine')
        self.assertEqual('Alanine', aa.name)
        self.assertEqual('Ala', aa.abbrev3)
        self.assertEqual('A', aa.abbrev1)
        self.assertEqual(('GCA', 'GCC', 'GCG', 'GCT'), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                'aliphaticity': 0.305785123967,
                'aromaticity': -0.550128534704,
                'composition': -1.0,
                'hydrogenation': 0.8973042362,
                'hydropathy': 0.4,
                'hydroxythiolation': -0.265160523187,
                'iep': -0.191489361702,
                'polar requirement': -0.463414634146,
                'polarity': -0.20987654321,
                'volume': -0.664670658683,
            },
            aa.propertyDetails)
        self.assertEqual(
            {
                'aliphaticity': 1,
                'aromaticity': 1,
                'composition': 1,
                'hydrogenation': 1,
                'hydropathy': 3,
                'hydroxythiolation': 2,
                'iep': 2,
                'polar requirement': 2,
                'polarity': 2,
                'volume': 2,
            },
            aa.propertyClusters)

    def testFindByNameCaseIgnored(self):
        """
        It must be possible to find an amino acid by its name when the name is
        given in mixed case.
        """
        aa = find('alaNIne')
        self.assertEqual('Alanine', aa.name)
        self.assertEqual('Ala', aa.abbrev3)
        self.assertEqual('A', aa.abbrev1)
        self.assertEqual(('GCA', 'GCC', 'GCG', 'GCT'), aa.codons)
        self.assertEqual(HYDROPHOBIC | SMALL | TINY, aa.properties)
        self.assertEqual(
            {
                'aliphaticity': 0.305785123967,
                'aromaticity': -0.550128534704,
                'composition': -1.0,
                'hydrogenation': 0.8973042362,
                'hydropathy': 0.4,
                'hydroxythiolation': -0.265160523187,
                'iep': -0.191489361702,
                'polar requirement': -0.463414634146,
                'polarity': -0.20987654321,
                'volume': -0.664670658683,
            },
            aa.propertyDetails)
        self.assertEqual(
            {
                'aliphaticity': 1,
                'aromaticity': 1,
                'composition': 1,
                'hydrogenation': 1,
                'hydropathy': 3,
                'hydroxythiolation': 2,
                'iep': 2,
                'polar requirement': 2,
                'polarity': 2,
                'volume': 2,
            },
            aa.propertyClusters)

    def testFindByNameCaseIgnoredNameWithSpace(self):
        """
        It must be possible to find an amino acid by its name when the name is
        given in mixed case, including if the name has a space in it.
        """
        aa = find('asparTIC aCId')
        self.assertEqual('Aspartic acid', aa.name)
        self.assertEqual('Asp', aa.abbrev3)
        self.assertEqual('D', aa.abbrev1)
        self.assertEqual(('GAC', 'GAT'), aa.codons)
        self.assertEqual(HYDROPHILIC | SMALL | POLAR | NEGATIVE, aa.properties)
        self.assertEqual(
            {
                'aliphaticity': -0.818181818182,
                'aromaticity': -1.0,
                'composition': 0.00363636363636,
                'hydrogenation': -0.90243902439,
                'hydropathy': -0.777777777778,
                'hydroxythiolation': -0.348394768133,
                'iep': -1.0,
                'polar requirement': 1.0,
                'polarity': 1.0,
                'volume': -0.389221556886,
            },
            aa.propertyDetails)
        self.assertEqual(
            {
                'aliphaticity': 1,
                'aromaticity': 1,
                'composition': 2,
                'hydrogenation': 1,
                'hydropathy': 1,
                'hydroxythiolation': 2,
                'iep': 1,
                'polar requirement': 4,
                'polarity': 4,
                'volume': 3,
            },
            aa.propertyClusters)


class TestPropertyClusters(TestCase):
    """
    Tests for the PROPERTY_CLUSTERS dict.
    """
    def testPropertyClusterKeys(self):
        """
        The PROPERTY_CLUSTERS dict must contain the right keys.
        """
        self.assertEqual(AA_LETTERS, sorted(PROPERTY_CLUSTERS.keys()))

    def testNumberOfValues(self):
        """
        Each key in PROPERTY_CLUSTERS must have a dict with 10 elements in it
        as the value.
        """
        for propertyNames in PROPERTY_CLUSTERS.values():
            self.assertEqual([
                'aliphaticity', 'aromaticity', 'composition',
                'hydrogenation', 'hydropathy', 'hydroxythiolation',
                'iep', 'polar requirement', 'polarity', 'volume'],
                sorted(propertyNames.keys()))

    def testAliphaticity(self):
        """
        Aliphaticity must always be in cluster 1.
        """
        for propertiesDict in PROPERTY_CLUSTERS.values():
            self.assertEqual(1, propertiesDict['aliphaticity'])

    def testPermittedClustersOnly(self):
        """
        All cluster numbers must be in {1, 2, 3, 4, 5}.
        """
        for propertiesDict in PROPERTY_CLUSTERS.values():
            for clusterNr in propertiesDict.values():
                self.assertIn(clusterNr, {1, 2, 3, 4, 5})


class TestPropertyDetailsRaw(TestCase):
    """
    Tests for the PROPERTY_DETAILS_RAW dict.
    """
    def testPropertyDetailsRawKeys(self):
        """
        The PROPERTY_DETAILS_RAW dict must contain the right keys.
        """
        self.assertEqual(AA_LETTERS, sorted(PROPERTY_DETAILS_RAW.keys()))

    def testNumberOfValues(self):
        """
        Each key in PROPERTY_DETAILS_RAW must have a dict with 10 elements in
        it as the value.
        """
        for propertyNames in PROPERTY_DETAILS_RAW.values():
            self.assertEqual([
                'aliphaticity', 'aromaticity', 'composition',
                'hydrogenation', 'hydropathy', 'hydroxythiolation',
                'iep', 'polar requirement', 'polarity', 'volume'],
                sorted(propertyNames))

    def testAliphaticity(self):
        """
        Aliphaticity of Alanin must have the right value.
        """
        self.assertEqual(0.239, PROPERTY_DETAILS_RAW['A']['aliphaticity'])

    def testValuesMustBeFloats(self):
        """
        Each key in PROPERTY_DETAILS_RAW must have a dict whose values are
        all floats.
        """
        for properties in PROPERTY_DETAILS_RAW.values():
            self.assertTrue(all(type(v) is float for v in properties.values()))


class TestPropertiesForSequence(TestCase):
    """
    Tests for the propertiesForSequence function in aa.py
    """
    def testUnknownProperty(self):
        """
        A C{ValueError} must be raised if an unknown property name is passed.
        """
        error = 'Unknown property: xxx'
        read = AARead('id', 'RRR')
        six.assertRaisesRegex(self, ValueError, error,
                              propertiesForSequence, read, ['xxx'])

    def testNoProperties(self):
        """
        If no properties are wanted, an empty dict must be returned.
        """
        read = AARead('id', 'RRR')
        self.assertEqual({}, propertiesForSequence(read, []))

    def testOnePropertyEmptySequence(self):
        """
        If one property is wanted but the sequence is empty, a dict with the
        property must be returned, and have an empty list value.
        """
        read = AARead('id', '')
        self.assertEqual(
            {
                'hydropathy': [],
            },
            propertiesForSequence(read, ['hydropathy']))

    def testPropertyNameIsLowercased(self):
        """
        The property name must be lower-cased in the result.
        """
        read = AARead('id', '')
        self.assertTrue('hydropathy' in
                        propertiesForSequence(read, ['HYDROPATHY']))

    def testOneProperty(self):
        """
        If one property is wanted, a dict with the property must be returned,
        and have the expected property values.
        """
        read = AARead('id', 'AI')
        self.assertEqual(
            {
                'hydropathy': [0.4, 1.0],
            },
            propertiesForSequence(read, ['hydropathy']))

    def testTwoProperties(self):
        """
        If two properties are wanted, a dict with the properties must be
        returned, and have the expected property values.
        """
        read = AARead('id', 'AI')
        self.assertEqual(
            {
                'composition': [-1.0, -1.0],
                'hydropathy': [0.4, 1.0],
            },
            propertiesForSequence(read, ['composition', 'hydropathy']))

    def testDuplicatedPropertyName(self):
        """
        If a property name is mentioned more than once, a dict with the
        wanted properties must be returned, and have the expected property
        values.
        """
        read = AARead('id', 'AI')
        self.assertEqual(
            {
                'composition': [-1.0, -1.0],
                'hydropathy': [0.4, 1.0],
            },
            propertiesForSequence(read, ['composition',
                                         'hydropathy', 'hydropathy']))

    def testMissingAminoAcid(self):
        """
        If an unknown amino acid appears in the sequence, its property value
        must be the default (-1.1).
        """
        read = AARead('id', 'XX')
        self.assertEqual(
            {
                'composition': [-1.1, -1.1],
                'hydropathy': [-1.1, -1.1],
            },
            propertiesForSequence(read, ['composition', 'hydropathy']))

    def testMissingAminoAcidWithNonDefaultMissingValue(self):
        """
        If an unknown amino acid appears in the sequence, its property value
        must be the missing AA value that was passed.
        """
        read = AARead('id', 'XX')
        self.assertEqual(
            {
                'composition': [-1.5, -1.5],
                'hydropathy': [-1.5, -1.5],
            },
            propertiesForSequence(read, ['composition', 'hydropathy'],
                                  missingAAValue=-1.5))


class TestClustersForSequence(TestCase):
    """
    Tests for the clustersForSequence function in aa.py
    """
    def testUnknownProperty(self):
        """
        A C{ValueError} must be raised if an unknown property name is passed.
        """
        error = 'Unknown property: xxx'
        read = AARead('id', 'RRR')
        six.assertRaisesRegex(self, ValueError, error,
                              clustersForSequence, read, ['xxx'])

    def testNoProperties(self):
        """
        If no properties are wanted, an empty dict must be returned.
        """
        read = AARead('id', 'RRR')
        self.assertEqual({}, clustersForSequence(read, []))

    def testOnePropertyEmptySequence(self):
        """
        If one property is wanted but the sequence is empty, a dict with the
        property must be returned, and have an empty list value.
        """
        read = AARead('id', '')
        self.assertEqual(
            {
                'hydropathy': [],
            },
            clustersForSequence(read, ['hydropathy']))

    def testPropertyNameIsLowercased(self):
        """
        The property name must be lower-cased in the result.
        """
        read = AARead('id', '')
        self.assertTrue('hydropathy' in
                        clustersForSequence(read, ['HYDROPATHY']))

    def testOneProperty(self):
        """
        If one property is wanted, a dict with the property must be returned,
        and have the expected property cluster number.
        """
        read = AARead('id', 'AI')
        self.assertEqual(
            {
                'hydropathy': [3, 4],
            },
            clustersForSequence(read, ['hydropathy']))

    def testTwoProperties(self):
        """
        If two properties are wanted, a dict with the properties must be
        returned, and have the expected property cluster numbers.
        """
        read = AARead('id', 'AI')
        self.assertEqual(
            {
                'composition': [1, 1],
                'hydropathy': [3, 4],
            },
            clustersForSequence(read, ['composition', 'hydropathy']))

    def testDuplicatedPropertyName(self):
        """
        If a property name is mentioned more than once, a dict with the
        wanted properties must be returned, and have the expected property
        cluster numbers.
        """
        read = AARead('id', 'AI')
        self.assertEqual(
            {
                'composition': [1, 1],
                'hydropathy': [3, 4],
            },
            clustersForSequence(read, ['composition',
                                       'hydropathy', 'hydropathy']))

    def testMissingAminoAcid(self):
        """
        If an unknown amino acid appears in the sequence, its property cluster
        must be the default (0).
        """
        read = AARead('id', 'XX')
        self.assertEqual(
            {
                'composition': [0, 0],
                'hydropathy': [0, 0],
            },
            clustersForSequence(read, ['composition', 'hydropathy']))

    def testMissingAminoAcidWithNonDefaultMissingValue(self):
        """
        If an unknown amino acid appears in the sequence, its property cluster
        must be the missing AA value that was passed.
        """
        read = AARead('id', 'XX')
        self.assertEqual(
            {
                'composition': [10, 10],
                'hydropathy': [10, 10],
            },
            clustersForSequence(read, ['composition', 'hydropathy'],
                                missingAAValue=10))


class TestCompareAaReads(TestCase):
    """
    Test the compareAaReads function.
    """
    def testEmptySequences(self):
        """
        Two empty sequences must compare as expected.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 0,
                    'gapMismatchCount': 0,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', ''),
                           AARead('id2', '')))

    def testExactMatch(self):
        """
        Two sequences that match exactly must compare as expected.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 5,
                    'gapMismatchCount': 0,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', 'GALHN'),
                           AARead('id2', 'GALHN')))

    def testMismatch(self):
        """
        If the sequences have mismatched bases, their count
        must be given correctly in the nonGapMismatchCount.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 3,
                    'gapMismatchCount': 0,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 2,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', 'GALYY'),
                           AARead('id2', 'GALHN')))

    def testNoOffsets(self):
        """
        If an empty set of wanted offsets is passed, the result must be empty.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 0,
                    'gapMismatchCount': 0,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', 'GAL-N'),
                           AARead('id2', 'G-LHN'), offsets=set()))

    def testOffsets(self):
        """
        If a set of wanted offsets is passed, the result must be restricted to
        just those offsets.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 1,
                    'gapMismatchCount': 0,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 1,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', 'GAL-L'),
                           AARead('id2', 'G-LHN'), offsets=set([0, 4])))

    def testGapInFirst(self):
        """
        A gap in the first sequence must be dealt with correctly.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 4,
                    'gapMismatchCount': 1,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [3],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', 'GAL-N'),
                           AARead('id2', 'GALHN')))

    def testGapInSecond(self):
        """
        A gap in the second sequence must be dealt with correctly.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 3,
                    'gapMismatchCount': 2,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [1, 2],
                },
            },
            compareAaReads(AARead('id1', 'GALHN'),
                           AARead('id2', 'G--HN')))

    def testNonDefaultGapChars(self):
        """
        We must be able to specify the gap characters.
        """
        for gap in '+$':
            self.assertEqual(
                {
                    'match': {
                        'matchCount': 3,
                        'gapMismatchCount': 2,
                        'gapGapMismatchCount': 0,
                        'nonGapMismatchCount': 0,
                    },
                    'read1': {
                        'extraCount': 0,
                        'gapOffsets': [2],
                    },
                    'read2': {
                        'extraCount': 0,
                        'gapOffsets': [0],
                    },
                },
                compareAaReads(AARead('id1', 'GA%sHN' % gap),
                               AARead('id2', '%sALHN' % gap), gapChars='+$'))

    def testGapGap(self):
        """
        Coinciding gaps in the sequences must be dealt with correctly
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 2,
                    'gapMismatchCount': 2,
                    'gapGapMismatchCount': 1,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [2, 3],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [1, 2],
                },
            },
            compareAaReads(AARead('id1', 'GA--N'),
                           AARead('id2', 'G--HN')))

    def testExtraInFirst(self):
        """
        If the first sequence has extra bases, they must be indicated in the
        extraCount.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 5,
                    'gapMismatchCount': 0,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 2,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', 'GALHNHN'),
                           AARead('id2', 'GALHN')))

    def testExtraInSecond(self):
        """
        If the second sequence has extra bases, they must be indicated in the
        extraCount.
        """
        self.assertEqual(
            {
                'match': {
                    'matchCount': 5,
                    'gapMismatchCount': 0,
                    'gapGapMismatchCount': 0,
                    'nonGapMismatchCount': 0,
                },
                'read1': {
                    'extraCount': 0,
                    'gapOffsets': [],
                },
                'read2': {
                    'extraCount': 2,
                    'gapOffsets': [],
                },
            },
            compareAaReads(AARead('id1', 'GALHN'),
                           AARead('id2', 'GALHNHN')))


class TestMatchToString(TestCase):
    """
    Test the matchToString function.
    """
    def testMismatchAndMatch(self):
        """
        Two sequences containing matches and mismatches must compare as
        expected.
        """
        read1 = AARead('id1', 'GALHNG')
        read2 = AARead('id2', 'GALHNA')
        match = compareAaReads(read1, read2)

        self.assertEqual(
            '''\
Matches: 5/6 (83.33%)
Mismatches: 1/6 (16.67%)
  Not involving gaps (i.e., conflicts): 1/6 (16.67%)
  Involving a gap in one sequence: 0
  Involving a gap in both sequences: 0
  Id: id1
    Length: 6
    Gaps: 0
  Id: id2
    Length: 6
    Gaps: 0''',
            matchToString(match, read1, read2)
        )

    def testGapAndMatch(self):
        """
        Two sequences containing matches and gaps must compare as
        expected.
        """
        read1 = AARead('id1', 'GALHN-')
        read2 = AARead('id2', 'GALHNA')
        match = compareAaReads(read1, read2)

        self.assertEqual(
            '''\
Matches: 5/6 (83.33%)
Mismatches: 1/6 (16.67%)
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 1/6 (16.67%)
  Involving a gap in both sequences: 0
  Id: id1
    Length: 6
    Gaps: 1/6 (16.67%)
    Gap locations (1-based): 6
  Id: id2
    Length: 6
    Gaps: 0''',
            matchToString(match, read1, read2)
        )



from collections import defaultdict


def findDistance(co1, co2):
    """
    Find the distance between two codons.

    @param co1: A C{str} of length three.
    @param co2: A C{str} of length three.
    @return: An C{int} distance in [0, 3].
    """
    return (co1[0] != co2[0]) + (co1[1] != co2[1]) + (co1[2] != co2[2])


def codonInformation(codons1, codons2):
    """
    Take two C{list} of codons, and returns information about the min and
    max distance between them.

    @param codon1: a C{list} of codons.
    @param codon2: a C{list} of codons.

    @return: a dict whose keys are C{int} distances and whose values are lists
        of pairs of codons.
    """
    result = defaultdict(list)
    for c1 in codons1:
        for c2 in codons2:
            distance = findDistance(c1, c2)
            result[distance].append([c1, c2])

    return result
